/* PAM : user account making module

   This module will create the user account if it doesn't exist.
   This allows to log in absolutely new users.

   Here is a sample /etc/pam.d/login file for Debian GNU/Linux
   2.1:

   auth       requisite  pam_securetty.so
   auth       sufficient pam_ldap.so
   auth       required   pam_unix.so
   auth       optional   pam_group.so
   auth       optional   pam_mail.so
   account    optional   pam_mkuser.so shell=/bin/bash
   account    requisite  pam_time.so
   account    sufficient pam_ldap.so
   account    required   pam_unix.so
   session    requires   pam_mkhomedir.so skel=/etc/skel/ umask=0077
   session    required   pam_unix.so
   session    optional   pam_lastlog.so
   password   required   pam_unix.so

   Released under the GNU LGPL version 2 or later.
   Copyright (c) BaseALT 2017.
   Based on the pam_mkhomedir by RedHat, originally written by
   Jason Gunthorpe <jgg@debian.org> Feb 1999
   Structure taken from pam_lastlogin by Andrew Morgan
   <morgan@parc.power.net> 1996.
 */

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>

/*
 * Here, we make a definition for the externally accessible function
 * in this file (this definition is required for static module
 * but strongly encouraged generally) it is used to instruct the
 * modules include the file to define the function prototypes.
 */

#define PAM_SM_SESSION

#include <security/pam_modules.h>
#include <security/_pam_macros.h>
#include <security/pam_modutil.h>
#include <security/pam_ext.h>

/* argument parsing */
#define MKUSER_DEBUG      020	/* be verbose about things */
#define MKUSER_QUIET      040	/* keep quiet about things */

struct options_t {
  int ctrl;
  const char *shell;
};
typedef struct options_t options_t;

static void
_pam_parse (const pam_handle_t *pamh, int flags, int argc, const char **argv,
            options_t *opt)
{
   opt->ctrl = 0;
   opt->shell = "/bin/sh";

   /* does the appliction require quiet? */
   if ((flags & PAM_SILENT) == PAM_SILENT)
      opt->ctrl |= MKUSER_QUIET;

   /* step through arguments */
   for (; argc-- > 0; ++argv)
   {
       if (!strcmp(*argv, "silent")) {
           opt->ctrl |= MKUSER_QUIET;
       } else if (!strcmp(*argv, "debug")) {
           opt->ctrl |= MKUSER_DEBUG;
       } else if (!strncmp(*argv,"shell=", 6)) {
           opt->shell = *argv + 6;
       } else {
           pam_syslog(pamh, LOG_ERR, "unknown option: %s", *argv);
       }
   }
}

/* Call the helper program to do the actual work of creating
the user account. */
static int
create_user (pam_handle_t *pamh, options_t *opt, const char *user)
{
   int retval, child;
   struct sigaction newsa, oldsa;

   /* Mention what is happening, if the notification fails that is OK */
   if (!(opt->ctrl & MKUSER_QUIET))
       pam_info(pamh, _("Creating user '%s'..."), user);


   D(("called."));

   /*
    * This code arranges that the demise of the child does not cause
    * the application to receive a signal it is not expecting - which
    * may kill the application or worse.
    */
   memset(&newsa, '\0', sizeof(newsa));
   newsa.sa_handler = SIG_DFL;
   sigaction(SIGCHLD, &newsa, &oldsa);

   if (opt->ctrl & MKUSER_DEBUG) {
        pam_syslog(pamh, LOG_DEBUG, "Executing 'useradd' command...");
   }

   /* Fork */
   child = fork();
   
   if (child == 0) {
       static char *envp[] = { NULL };
       const char *args[] = { NULL, NULL, NULL, NULL };

       if (pam_modutil_sanitize_helper_fds(pamh,
                                           PAM_MODUTIL_NULL_FD,
                                           PAM_MODUTIL_NULL_FD,
                                           PAM_MODUTIL_NULL_FD) < 0)
       {
           _exit(PAM_SYSTEM_ERR);
       }

       /* Execute the useradd command */
       args[0] = USERADD_COMMAND;
       args[1] = "-s";
       args[2] = opt->shell;
       args[3] = user;
    
       execve(USERADD_COMMAND, (char *const *) args, envp);

       /* Should not get here! Exit with an error. */
       D(("useradd binary is not available"));
       _exit(PAM_SYSTEM_ERR);
   } else if (child > 0) {
       int rc;
       while ((rc = waitpid(child, &retval, 0)) < 0 && errno == EINTR);
       if (rc < 0) {
           pam_syslog(pamh, LOG_ERR, "waitpid failed: %m");
           retval = PAM_SYSTEM_ERR;
       } else if (!WIFEXITED(retval)) {
           pam_syslog(pamh, LOG_ERR, "useradd abnormal exit: %d", retval);
           retval = PAM_SYSTEM_ERR;
       } else {
           retval = WEXITSTATUS(retval);
           if (opt->ctrl & MKUSER_DEBUG) {
               pam_syslog(pamh, LOG_DEBUG, "useradd returned %d", retval);
           }
           if (0 == retval) {
               retval = PAM_SUCCESS;
           } else {
               retval = PAM_PERM_DENIED;
           }
       }
   } else {
       D(("fork failed"));
       pam_syslog(pamh, LOG_ERR, "fork failed: %m");
       retval = PAM_SYSTEM_ERR;
   }

   sigaction(SIGCHLD, &oldsa, NULL);   /* restore old signal handler */

   if (retval != PAM_SUCCESS && !(opt->ctrl & MKUSER_QUIET)) {
       pam_error(pamh, _("Unable to create user '%s'."), user);
   }

   D(("returning %d", retval));
   return retval;
}

/* --- Authentication management functions (only) --- */

int
pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
                     const char **argv)
{
   int retval;
   options_t opt;
   const char *user = NULL;
   const struct passwd *pwd = NULL;
   struct stat St;

   /* Parse the flag values */
   _pam_parse(pamh, flags, argc, argv, &opt);

   /* Determine the user name  */
   const void *_user = NULL;
   retval = pam_get_item(pamh, PAM_USER, &_user);
   if (retval != PAM_SUCCESS || _user == NULL || *(const char *) _user == '\0')
   {
       pam_syslog(pamh, LOG_NOTICE, "Cannot obtain the user name.");
       return PAM_CRED_INSUFFICIENT;
   }
   user = (const char *) _user;
   
   /* Get the password entry */
   pwd = pam_modutil_getpwnam(pamh, user);
   if (NULL == pwd)
   {
       pam_syslog(pamh, LOG_NOTICE, "User is unknown. Will create the account");
   } else {
       pam_syslog(pamh, LOG_DEBUG, "User is known. Nothing to do");
       return PAM_SUCCESS;
   }

   return create_user(pamh, &opt, user);
}

/* Ignore */
int pam_sm_close_session (pam_handle_t * pamh, int flags,
                          int argc, const char **argv)
{
   return PAM_SUCCESS;
}

#if defined __linux__ && defined __ELF__
__asm__(".globl pam_sm_acct_mgmt; pam_sm_acct_mgmt = pam_sm_open_session");
#else
int
pam_sm_acct_mgmt (pam_handle_t *pamh, int flags,
		  int argc, const char **argv)
{
	return pam_sm_open_session(pamh, flags, argc, argv);
}
#endif
